@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004 Frank Denis. All rights reserved.
+ * Copyright (c) 2004-2007 Frank Denis. All rights reserved.
  * 
  * This crucial part of UCARP is derived from the OpenBSD project.
  * Original copyright follows. 
@@ -55,20 +55,21 @@ static void carp_set_state(struct carp_softc *sc, int state)
     }
     switch (state) {
     case INIT:
-        logfile(LOG_INFO, _("Switching to state : INIT"));
+        logfile(LOG_INFO, _("Switching to state: INIT"));
         break;
     case BACKUP:
-        logfile(LOG_WARNING, _("Switching to state : BACKUP"));
-        (void) spawn_handler(dev_desc_fd, downscript);
-        gratuitous_arp(dev_desc_fd, 0);
+        logfile(LOG_WARNING, _("Switching to state: BACKUP"));
+        if ((sc->sc_state != INIT) || (neutral != 1)) {
+            (void) spawn_handler(dev_desc_fd, downscript);
+        }
         break;
     case MASTER:
-        logfile(LOG_WARNING, _("Switching to state : MASTER"));
+        logfile(LOG_WARNING, _("Switching to state: MASTER"));
         (void) spawn_handler(dev_desc_fd, upscript);
-        gratuitous_arp(dev_desc_fd, 1);
+        gratuitous_arp(dev_desc_fd);        
         break;
     default:
-        logfile(LOG_ERR, _("Unknown state : [%d]"), (int) state);
+        logfile(LOG_ERR, _("Unknown state: [%d]"), (int) state);
         abort();
     }
     sc->sc_state = state;
@@ -175,16 +176,15 @@ static void carp_send_ad(struct carp_softc *sc)
     struct carp_header ch;
     struct ether_header eh;
     struct timeval tv;
-    struct carp_header *ch_ptr;
-    struct ether_header *eh_ptr;
     struct ip ip;
-    struct ip *ip_ptr;
+    unsigned char *ip_ptr;
     unsigned char *pkt;
     unsigned short sum;    
     size_t ip_len;
     size_t eth_len;
     int advbase;
     int advskew;
+    int rc;
 
 #ifdef DEBUG
     logfile(LOG_DEBUG, "-> carp_send_ad()");
@@ -226,45 +226,45 @@ static void carp_send_ad(struct carp_softc *sc)
     ip.ip_p = IPPROTO_CARP;
     ip.ip_sum = 0;
     
-    memcpy(&ip.ip_src, &srcip, sizeof ip.ip_src);
+    memcpy(&ip.ip_src, &srcip, sizeof ip.ip_src);    
+    memcpy(&ip.ip_dst.s_addr, inaddr_carp_group, sizeof ip.ip_dst.s_addr);
     
-    ip.ip_dst.s_addr = INADDR_CARP_GROUP;
+    carp_prepare_ad(&ch, sc);       
+    
+    ch.carp_cksum = 0;
+    sum = cksum(&ch, sizeof ch);
+    ch.carp_cksum = htons(sum);    
+    
+    eh.ether_shost[0] = 0x00;
+    eh.ether_shost[1] = 0x00;
+    eh.ether_shost[2] = 0x5e;
+    eh.ether_shost[3] = 0x00;
+    eh.ether_shost[4] = 0x00;
+    eh.ether_shost[5] = vhid;
+    
+    eh.ether_dhost[0] = 0x01;
+    eh.ether_dhost[1] = 0x00;
+    eh.ether_dhost[2] = 0x5e;
+    eh.ether_dhost[3] = 0x00;
+    eh.ether_dhost[4] = 0x00;
+    eh.ether_dhost[5] = 0x12;        
+    
+    eh.ether_type = htons(ETHERTYPE_IP);    
     
     memcpy(pkt, &eh, sizeof eh);
     memcpy(pkt + sizeof eh, &ip, sizeof ip);
     memcpy(pkt + sizeof ip + sizeof eh, &ch, sizeof ch);
-    eh_ptr = (struct ether_header *) pkt;
-    ip_ptr = (struct ip *) (pkt + sizeof eh);        
-    ch_ptr = (struct carp_header *) (pkt + sizeof eh + sizeof ip);
-    
-    carp_prepare_ad(ch_ptr, sc);       
-    
-    ch_ptr->carp_cksum = 0;
-    sum = cksum(ch_ptr, sizeof ch);
-    ch_ptr->carp_cksum = htons(sum);
+
+    ip_ptr = pkt + sizeof eh;
     sum = cksum(ip_ptr, ip_len);
-    ip_ptr->ip_sum = htons(sum);
-    
-    eh_ptr->ether_shost[0] = 0x00;
-    eh_ptr->ether_shost[1] = 0x00;
-    eh_ptr->ether_shost[2] = 0x5e;
-    eh_ptr->ether_shost[3] = 0x00;
-    eh_ptr->ether_shost[4] = 0x00;
-    eh_ptr->ether_shost[5] = vhid;
-    
-    eh_ptr->ether_dhost[0] = 0xff;
-    eh_ptr->ether_dhost[1] = 0xff;
-    eh_ptr->ether_dhost[2] = 0xff;
-    eh_ptr->ether_dhost[3] = 0xff;
-    eh_ptr->ether_dhost[4] = 0xff;
-    eh_ptr->ether_dhost[5] = 0xff;        
-    
-    eh_ptr->ether_type = htons(ETHERTYPE_IP);
-    
-    if (write(dev_desc_fd, pkt, eth_len) < 0) {
-#ifdef DEBUG
-        logfile(LOG_WARNING, _("write() has failed"));
-#endif
+    ip_ptr[offsetof(struct ip, ip_sum)] = (sum >> 8) & 0xff;
+    ip_ptr[offsetof(struct ip, ip_sum) + 1] = sum & 0xff;
+
+    do {
+        rc = write(dev_desc_fd, pkt, eth_len);
+    } while (rc < 0 && errno == EINTR);
+    if (rc < 0) {
+        logfile(LOG_WARNING, _("write() has failed: %s"), strerror(errno));
         if (sc->sc_sendad_errors < INT_MAX) {
             sc->sc_sendad_errors++;
         }
@@ -295,6 +295,15 @@ static void carp_send_ad(struct carp_softc *sc)
 #endif
     
     ALLOCA_FREE(pkt);
+    
+    if (sc->sc_delayed_arp > 0)
+        sc->sc_delayed_arp--;
+    if (sc->sc_delayed_arp == 0) {
+        if (sc->sc_state == MASTER) {
+            gratuitous_arp(dev_desc_fd);
+        }
+        sc->sc_delayed_arp = -1;
+    }    
     if (advbase != 255 || advskew != 255) {
         sc->sc_ad_tmo.tv_sec = now.tv_sec + tv.tv_sec;
         sc->sc_ad_tmo.tv_usec = now.tv_usec + tv.tv_usec;            
@@ -303,7 +312,7 @@ static void carp_send_ad(struct carp_softc *sc)
 }
 
 static void carp_send_ad_all(struct carp_softc *sc) {
-    return carp_send_ad(sc);
+    carp_send_ad(sc);
 }
 
 static void carp_setrun(struct carp_softc *sc, sa_family_t af)
@@ -313,18 +322,14 @@ static void carp_setrun(struct carp_softc *sc, sa_family_t af)
 #ifdef DEBUG
     logfile(LOG_DEBUG, "carp_setrun()");
 #endif
+    if (gettimeofday(&now, NULL) != 0) {
+        logfile(LOG_WARNING, _("initializing now to gettimeofday() failed: %s"),
+                strerror(errno));
+    }
     switch (sc->sc_state) {
     case INIT:
-        if (preempt != 0 && carp_suppress_preempt == 0) {
-            carp_send_ad(sc);
-            gratuitous_arp(dev_desc_fd, 1);
-#ifdef INET6
-            carp_send_na(sc);
-#endif /* INET6 */
-            carp_set_state(sc, MASTER);
-        } else {
-            carp_set_state(sc, BACKUP);
-        }
+        carp_set_state(sc, BACKUP);
+        carp_setrun(sc, 0);         
         break;
     case BACKUP:
         sc->sc_ad_tmo.tv_sec = 0;
@@ -342,14 +347,12 @@ static void carp_setrun(struct carp_softc *sc, sa_family_t af)
             break;
 #endif /* INET6 */
         default:
-            if (sc->sc_md_tmo.tv_sec != 0) {
-                sc->sc_md_tmo.tv_sec = now.tv_sec + tv.tv_sec;
-                sc->sc_md_tmo.tv_usec = now.tv_usec + tv.tv_usec;
-            }
-            if (sc->sc_md6_tmo.tv_sec != 0) {
-                sc->sc_md6_tmo.tv_sec = now.tv_sec + tv.tv_sec;
-                sc->sc_md6_tmo.tv_usec = now.tv_usec + tv.tv_usec;
-            }
+            sc->sc_md_tmo.tv_sec = now.tv_sec + tv.tv_sec;
+            sc->sc_md_tmo.tv_usec = now.tv_usec + tv.tv_usec;
+#ifdef INET6
+            sc->sc_md6_tmo.tv_sec = now.tv_sec + tv.tv_sec;
+            sc->sc_md6_tmo.tv_usec = now.tv_usec + tv.tv_usec;
+#endif
             break;
         }        
         break;
@@ -379,7 +382,8 @@ static void carp_master_down(struct carp_softc *sc)
     case BACKUP:
         carp_set_state(sc, MASTER);
         carp_send_ad(sc);
-        gratuitous_arp(dev_desc_fd, 0);
+        /* Schedule a delayed ARP request to deal w/ some L3 switches */
+        sc->sc_delayed_arp = 2;
 #ifdef INET6
         carp_send_na(sc);
 #endif /* INET6 */
@@ -401,7 +405,7 @@ static void packethandler(unsigned char *dummy,
     unsigned int ip_len;
             
     (void) dummy;
-    if (header->caplen < (sizeof etherhead + sizeof iphead)) {
+    if (header->caplen <= (sizeof etherhead + sizeof iphead)) {
         return;
     }    
     memcpy(&etherhead, sp, sizeof etherhead);
@@ -426,10 +430,13 @@ static void packethandler(unsigned char *dummy,
     sp += sizeof etherhead;
     caplen = header->caplen - sizeof etherhead;
     memcpy(&iphead, sp, sizeof iphead);    
+    if (iphead.ip_src.s_addr == srcip.s_addr) {
+        return;
+    }
     ip_len = iphead.ip_hl << 2;
     source = ntohl(iphead.ip_src.s_addr);
     dest = ntohl(iphead.ip_dst.s_addr);
-    proto = iphead.ip_p;    
+    proto = iphead.ip_p;
     switch (proto) {
     case IPPROTO_CARP: {
         struct carp_header ch;
@@ -446,10 +453,10 @@ static void packethandler(unsigned char *dummy,
                 dest >> 24 & 0xff, dest >> 16 & 0xff,
                 dest >> 8 & 0xff, dest & 0xff);
 #endif
-        if (caplen != ip_len + sizeof ch) {
+        if (caplen < ip_len + sizeof ch) {
 #ifdef DEBUG
             logfile(LOG_DEBUG,
-                    "Bogus size : caplen=[%u], ip_len=[%u] ch_len=[%u]",
+                    "Bogus size: caplen=[%u], ip_len=[%u] ch_len=[%u]",
                     (unsigned int) caplen, (unsigned int) ip_len,
                     (unsigned int) sizeof ch);
 #endif
@@ -458,37 +465,37 @@ static void packethandler(unsigned char *dummy,
         memcpy(&ch, sp + ip_len, sizeof ch);
 #ifdef EXTRA_DEBUG
         logfile(LOG_DEBUG,
-                "Capture len : [%u] carp_header : [%u] ip_header : [%u]",
+                "Capture len: [%u] carp_header: [%u] ip_header: [%u]",
                 (unsigned int) caplen, (unsigned int) sizeof ch,
                 (unsigned int) ip_len);
-        logfile(LOG_DEBUG, "CARP type : [%u] version : [%u]",
+        logfile(LOG_DEBUG, "CARP type: [%u] version: [%u]",
                 (unsigned) ch.carp_type, (unsigned) ch.carp_version);
-        logfile(LOG_DEBUG, "CARP vhid : [%u] advskew : [%u]",
+        logfile(LOG_DEBUG, "CARP vhid: [%u] advskew: [%u]",
                 (unsigned) ch.carp_vhid, (unsigned) ch.carp_advskew);
-        logfile(LOG_DEBUG, "CARP advbase : [%u] cksum : [%u]",
+        logfile(LOG_DEBUG, "CARP advbase: [%u] cksum: [%u]",
                 (unsigned) ch.carp_advbase, (unsigned) ch.carp_cksum);
-        logfile(LOG_DEBUG, "CARP counter : [%lu][%lu]",
+        logfile(LOG_DEBUG, "CARP counter: [%lu][%lu]",
                 (unsigned long) ch.carp_counter[0],
                 (unsigned long) ch.carp_counter[1]);
 #endif
         if (iphead.ip_ttl != CARP_DFLTTL) {
-            logfile(LOG_WARNING, _("Bad TTL : [%u]"),                    
+            logfile(LOG_WARNING, _("Bad TTL: [%u]"),                    
                     (unsigned int) iphead.ip_ttl);
             return;
         }        
         if (ch.carp_version != CARP_VERSION) {
-            logfile(LOG_WARNING, _("Bad version : [%u]"),
+            logfile(LOG_WARNING, _("Bad version: [%u]"),
                     (unsigned int) ch.carp_version);
             return;
         }
         if (ch.carp_vhid != vhid) {
 #ifdef DEBUG
-            logfile(LOG_NOTICE, _("Ignoring vhid : [%u]"),
+            logfile(LOG_NOTICE, _("Ignoring vhid: [%u]"),
                     (unsigned int) ch.carp_vhid);
 #endif
             return;            
         }
-        if (cksum(sp, caplen) != 0) {
+        if (cksum(sp, ip_len + sizeof ch) != 0) {
             logfile(LOG_WARNING, _("Bad IP checksum"));
             return;
         }        
@@ -555,15 +562,32 @@ static void packethandler(unsigned char *dummy,
         case MASTER:
             /*
              * If we receive an advertisement from a master who's going to
-             * be more frequent than us, go into BACKUP state.
+             * be more frequent than us, or from a master who's advertising
+             * with the same frequency as us but with a lower IP address, 
+             * go into BACKUP state.
              */
             if (timercmp(&sc_tv, &ch_tv, >) ||
-                timercmp(&sc_tv, &ch_tv, ==)) {
-                sc.sc_ad_tmo.tv_sec = 0;
+                (timercmp(&sc_tv, &ch_tv, ==) &&
+                 iphead.ip_src.s_addr < srcip.s_addr)) {
                 carp_set_state(&sc, BACKUP);
                 carp_setrun(&sc, 0);
-                logfile(LOG_WARNING, _("Going back to BACKUP state"));
+                logfile(LOG_WARNING, _("Preferred master advertised: "
+                                       "going back to BACKUP state"));
             }
+
+            /*
+             * If we receive an advertisement from a master who's advertising
+             * with the same frequency as us but with a higher IP address,
+             * reassert our dominance by issuing another gratuitous arp.
+             */
+            if (timercmp(&sc_tv, &ch_tv, ==) &&
+                iphead.ip_src.s_addr > srcip.s_addr) {
+                gratuitous_arp(dev_desc_fd);
+                sc.sc_delayed_arp = 2; /* and yet another in 2 ticks */
+                logfile(LOG_WARNING, _("Non-preferred master advertising: "
+                                       "reasserting control of VIP with another gratuitous arp"));
+            }
+
             break;
         case BACKUP:
             /*
@@ -610,22 +634,57 @@ RETSIGTYPE sighandler_exit(const int sig)
     logfile(LOG_DEBUG, "sighandler_exit(): Calling [%s] and exiting",
             downscript);
 #endif
-    (void) spawn_handler(dev_desc_fd, downscript);
+    if (sc.sc_state != BACKUP) {
+        (void) spawn_handler(dev_desc_fd, downscript);
+    }
     _exit(EXIT_SUCCESS);
 }
 
+RETSIGTYPE sighandler_usr(const int sig)
+{
+    switch (sig) {
+    case SIGUSR1:
+	received_signal=1;
+	break;
+    case SIGUSR2:
+	received_signal=2;
+	break;
+    }
+}
+
+char *build_bpf_rule(void)
+{
+    static char rule[256];
+    const char *srcip_s;    
+    
+    if ((srcip_s = inet_ntoa(srcip)) == NULL) {
+        logfile(LOG_ERR, "inet_ntoa: [%s]", strerror(errno));
+        return NULL;
+    }
+    snprintf(rule, sizeof rule, "proto %u and src host not %s",
+             (unsigned int) IPPROTO_CARP, srcip_s);
+#ifdef DEBUG
+    logfile(LOG_DEBUG, "BPF rule: [%s]", rule);
+#endif
+    
+    return rule;
+}
+
 int docarp(void)
 {
+    char errbuf[PCAP_ERRBUF_SIZE];
     struct bpf_program bpfp;
     struct pollfd pfds[1];
+    struct ifreq iface;
+    int fd;
     int nfds;
-    char errbuf[PCAP_ERRBUF_SIZE];
+    int iface_running = 1;
 
     sc.sc_vhid = vhid;
     sc.sc_advbase = advbase;
     sc.sc_advskew = advskew;
     sc.sc_init_counter = 1;
-    sc.sc_naddrs = sc.sc_naddrs6 = 0;
+    sc.sc_delayed_arp = -1;
 #ifdef INET6
     sc.sc_im6o.im6o_multicast_hlim = CARP_DFLTTL;
 #endif /* INET6 */
@@ -648,7 +707,7 @@ int docarp(void)
 
     if (fill_mac_address() != 0) {
         logfile(LOG_ERR, _("Unable to find MAC address of [%s]"),
-                interface);
+                interface == NULL ? "-" : interface);
         return -1;
     }
     logfile(LOG_INFO, _("Local advertised ethernet address is "
@@ -656,42 +715,110 @@ int docarp(void)
             (unsigned int) hwaddr[0], (unsigned int) hwaddr[1],
             (unsigned int) hwaddr[2], (unsigned int) hwaddr[3],
             (unsigned int) hwaddr[4], (unsigned int) hwaddr[5]);
-    if ((dev_desc = pcap_open_live(interface, ETHERNET_MTU, 1,
+    if ((dev_desc = pcap_open_live(interface, ETHERNET_MTU, 0,
                                    CAPTURE_TIMEOUT, errbuf)) == NULL) {
-        logfile(LOG_ERR, _("Unable to open interface [%s] : %s"),
-                interface, errbuf);
+        logfile(LOG_ERR, _("Unable to open interface [%s]: %s"),
+                interface == NULL ? "-" : interface, errbuf);
         return -1;
     }    
-    if (pcap_compile(dev_desc, &bpfp, (char *) BPF_CARP_RULE,
+    if (pcap_compile(dev_desc, &bpfp, build_bpf_rule(),
                      1, (bpf_u_int32) 0) != 0) {
-        logfile(LOG_ERR, _("Unable to compile pcarp rule : %s [%s]"),
-                errbuf, interface);
+        logfile(LOG_ERR, _("Unable to compile pcap rule: %s [%s]"),
+                errbuf, interface == NULL ? "-" : interface);
         return -1;
     }
     pcap_setfilter(dev_desc, &bpfp);
     dev_desc_fd = pcap_fileno(dev_desc);
     pfds[0].fd = dev_desc_fd;
-    pfds[0].events = POLLIN | POLLERR | POLLHUP | POLLNVAL;
-    
-    carp_setrun(&sc, 0);
+    pfds[0].events = POLLIN | POLLERR | POLLHUP;
     
     if (shutdown_at_exit != 0) {
         (void) signal(SIGINT, sighandler_exit);
         (void) signal(SIGQUIT, sighandler_exit);
         (void) signal(SIGTERM, sighandler_exit);
     }
+    (void) signal(SIGUSR1, sighandler_usr);
+    (void) signal(SIGUSR2, sighandler_usr);
+
+    if (gettimeofday(&now, NULL) != 0) {
+        logfile(LOG_WARNING, _("initializing now to gettimeofday() failed: %s"),
+                strerror(errno));
+    }
+    carp_setrun(&sc, 0); 
     
+    if ((fd = socket(PF_INET, SOCK_DGRAM, 0)) == -1) {
+        logfile(LOG_ERR, _("Error opening socket for interface [%s]: %s"),
+                interface == NULL ? "-" : interface, strerror(errno));
+        return -1;
+    }
+#ifdef SIOCGIFFLAGS    
+    if (strlen(interface) >= sizeof iface.ifr_name) {
+        logfile(LOG_ERR, _("Interface name too long"));
+        return -1;
+    }
+    strncpy(iface.ifr_name, interface, sizeof iface.ifr_name);
+#endif    
     for (;;) {
-        nfds = poll(pfds, (nfds_t) 1, (int) (sc.sc_advbase * 1000));
-        if (nfds == -1 || 
-            (pfds[0].revents & (POLLERR | POLLHUP | POLLNVAL)) != 0) {
-#ifdef DEBUG
-            logfile(LOG_NOTICE, "poll() error");
+#ifdef SIOCGIFFLAGS
+        if (ioctl(fd, SIOCGIFFLAGS, &iface) != 0) {
+            break;
+        }
+        if ((iface.ifr_flags & IFF_RUNNING) == 0) {
+            carp_set_state(&sc, BACKUP);
+            sc.sc_ad_tmo.tv_sec = 0;
+            sc.sc_ad_tmo.tv_usec = 0;
+            sc.sc_md_tmo.tv_sec = 0;
+            sc.sc_md6_tmo.tv_usec = 0;
+            if (iface_running) {
+                iface_running = 0;
+            }
+            sleep(SECONDS_TO_WAIT_AFTER_INTERFACE_IS_DOWN);
+            continue;
+        } else {
+            if (!iface_running) {
+                iface_running = 1;
+                carp_setrun(&sc, 0);
+            }
+        }
 #endif
+        if (received_signal != 0) {
+            int flag = received_signal;
+	    
+            received_signal = 0;
+            switch (flag) {
+	    case 1:
+		logfile(LOG_INFO, "%s on %s id %d", 
+			(sc.sc_state == BACKUP ? "BACKUP" : "MASTER"),
+			interface, sc.sc_vhid);
+		break;
+	    case 2:
+#ifdef DEBUG
+		logfile(LOG_DEBUG, "Caught signal (USR2) considering going down");
+#endif
+		if (sc.sc_state != BACKUP) {
+		    carp_set_state(&sc, BACKUP);
+		    sleep(3); /* hold up a sec... */
+		    carp_setrun(&sc, 0); /* now listen for 3 heartbeats, as usual */
+		    continue;
+		}
+		break;
+            }
+        }
+	
+        nfds = poll(pfds, (nfds_t) 1, sc.sc_advbase * 1000);
+        if (nfds == -1) {
+            if (errno == EINTR) {
+                continue;
+            }
+            logfile(LOG_ERR, _("exiting: poll() error: %s"), strerror(errno));
+        }
+        if ((pfds[0].revents & (POLLERR | POLLHUP)) != 0) {
+            logfile(LOG_ERR, _("exiting: pfds[0].revents = %d"),
+                    pfds[0].revents);
             break;
         }
         if (gettimeofday(&now, NULL) != 0) {
-            logfile(LOG_WARNING, "gettimeofday() failed : %s",
+            logfile(LOG_WARNING, _("gettimeofday() failed: %s"),
                     strerror(errno));
             continue;
         }        
@@ -701,9 +828,11 @@ int docarp(void)
         if (sc.sc_md_tmo.tv_sec != 0 && timercmp(&now, &sc.sc_md_tmo, >)) {
             carp_master_down(&sc);
         }
+#ifdef INET6
         if (sc.sc_md6_tmo.tv_sec != 0 && timercmp(&now, &sc.sc_md6_tmo, >)) {
             carp_master_down(&sc);
         }
+#endif
         if (sc.sc_ad_tmo.tv_sec != 0 && timercmp(&now, &sc.sc_ad_tmo, >)) {
             carp_send_ad(&sc);
         }
